import numpy as np
import pandas as pd
import pytest
import rasterio
import xarray as xr
from numpy.testing import assert_allclose

import openamundsen as oa
import openamundsen.errors as errors
from openamundsen.fileio.griddedoutput import _freq_write_dates

from .conftest import base_config


def test_freq_write_dates():
    dates = pd.date_range(start="2021-01-01 00:00", end="2021-12-31 23:00", freq="h")

    assert _freq_write_dates(dates, "h", False).equals(dates)

    wd = _freq_write_dates(dates, "3h", False)
    assert np.all(wd.hour.isin([0, 3, 6, 9, 12, 15, 18, 21]))

    wd = _freq_write_dates(dates, "D", False)
    assert np.all(wd.hour == 0)
    assert dates.normalize().unique().equals(wd.normalize())

    wd = _freq_write_dates(dates, "D", True)
    assert np.all(wd.hour == 23)

    wd = _freq_write_dates(dates, "ME", True)
    assert np.all(wd.day.isin([28, 30, 31]))
    assert np.all(wd.hour == 23)

    assert _freq_write_dates(dates, "YE", False).equals(pd.to_datetime(["2021-12-31 23:00"]))
    assert _freq_write_dates(dates, "YS", False).equals(pd.to_datetime(["2021-01-01 00:00"]))

    wd = _freq_write_dates(dates, "ME", False)
    assert np.all(wd.day.isin([28, 30, 31]))
    assert np.all(wd.hour == 23)

    wd = _freq_write_dates(dates, "MS", False)
    assert np.all(wd.day == 1)
    assert np.all(wd.hour == 0)

    dates = pd.date_range(start="2021-01-01 06:00", end="2021-12-31 21:00", freq="3h")
    wd = _freq_write_dates(dates, "D", False)
    assert np.all(wd.hour == 6)

    wd = _freq_write_dates(dates, "D", True)
    assert np.all(wd.hour == 3)

    with pytest.raises(ValueError):
        wd = _freq_write_dates(dates, "h", False)

    with pytest.raises(ValueError):
        wd = _freq_write_dates(dates, "h", True)

    dates = pd.date_range(start="2021-01-01 02:00", end="2021-12-31 04:00", freq="3h")
    wd = _freq_write_dates(dates, "D", False)
    assert np.all(wd.hour == 2)
    # assert _freq_write_dates(dates, 'A', False).size == 0

    dates = pd.date_range(start="2021-01-01 02:00", end="2021-01-01 07:00", freq="h")
    assert len(_freq_write_dates(dates, "D", False)) == 1
    assert len(_freq_write_dates(dates, "D", True)) == 0

    dates = pd.date_range(start="2021-01-15 00:00", end="2021-03-15 21:00", freq="3h")
    wd = _freq_write_dates(dates, "ME", True)
    assert len(wd) == 2

    dates = pd.date_range(start="2021-01-01 01:00", end="2021-12-31 22:00", freq="3h")
    assert _freq_write_dates(dates, "YE", False).equals(pd.to_datetime(["2021-12-31 22:00"]))
    assert _freq_write_dates(dates, "YS", False).equals(pd.to_datetime(["2021-01-01 01:00"]))

    for freq in ("BME", "QE", "ms", "us", "ns", "foo"):
        with pytest.raises(errors.ConfigurationError):
            _freq_write_dates(dates, freq, False)


@pytest.mark.parametrize("fmt", ["netcdf", "ascii", "geotiff", "memory"])
def test_formats(fmt, tmp_path):
    # Skip test for format "ascii" for certain GDAL versions because of this bug:
    # https://github.com/OSGeo/gdal/issues/9666
    if fmt == "ascii" and rasterio.__gdal_version__ in (
        "3.8.0",
        "3.8.1",
        "3.8.2",
        "3.8.3",
        "3.8.4",
        "3.8.5",
    ):
        pytest.skip()

    config = base_config()
    config.end_date = "2020-01-16"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = fmt
    grid_cfg.variables = [
        {"var": "meteo.temp", "freq": "D"},
        {"var": "meteo.precip", "freq": "D", "agg": "sum"},
        {"var": "meteo.sw_in", "dates": ["2020-01-15 12:00"], "name": "myvar"},
        {"var": "base.dem", "dates": ["2020-01-15 00:00"]},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    if fmt in ("netcdf", "memory"):
        if fmt == "netcdf":
            ds = xr.open_dataset(tmp_path / "output_grids.nc")
        elif fmt == "memory":
            ds = model.gridded_output.data

        assert "temp" in ds.variables
        assert ds.time1.to_index().equals(
            pd.DatetimeIndex(
                [
                    "2020-01-15 00:00",
                    "2020-01-16 00:00",
                ]
            )
        )
        assert "precip" in ds.variables
        assert ds.time2.to_index().equals(
            pd.DatetimeIndex(
                [
                    "2020-01-15 21:00",
                    "2020-01-16 21:00",
                ]
            )
        )
        assert pd.DatetimeIndex(ds.time2_bounds.values.ravel()).equals(
            pd.DatetimeIndex(
                [
                    "2020-01-15 00:00:00",
                    "2020-01-15 21:00:00",
                    "2020-01-15 21:00:00",
                    "2020-01-16 21:00:00",
                ]
            )
        )
        assert "myvar" in ds.variables
        assert ds.time3.to_index().equals(pd.DatetimeIndex(["2020-01-15 12:00"]))
        assert_allclose(model.state.base.dem, ds.dem.loc["2020-01-15 00:00", :, :].values)
    else:
        if fmt == "ascii":
            ext = "asc"
        elif fmt == "geotiff":
            ext = "tif"

        fn = tmp_path / f"temp_2020-01-16T0000.{ext}"
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f"precip_2020-01-16T0000_2020-01-16T2100.{ext}"
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f"myvar_2020-01-15T1200.{ext}"
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f"dem_2020-01-15T0000.{ext}"
        with rasterio.open(fn) as ds:
            assert_allclose(model.state.base.dem, ds.read(1))


def test_duplicate_output_names(tmp_path):
    config = base_config()
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {"var": "meteo.temp"},
        {"var": "snow.temp"},
    ]

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()


def test_values(tmp_path):
    config = base_config()
    config.start_date = "2020-01-17"
    config.end_date = "2020-01-19"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {"var": "meteo.temp", "freq": "D", "agg": "mean"},
        {"var": "meteo.precip", "freq": "D", "agg": "sum"},
        {"var": "meteo.sw_in", "dates": ["2020-01-17 12:00"]},
        {"var": "snow.swe", "freq": "D"},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    data_vals = {
        "temp": {},
        "precip": {},
        "sw_in": {},
        "swe": {},
    }

    for date in model.dates:
        model.run_single()
        date = date.strftime("%Y-%m-%d %H:%M")
        data_vals["temp"][date] = model.state.meteo.temp.copy()
        data_vals["precip"][date] = model.state.meteo.precip.copy()
        data_vals["sw_in"][date] = model.state.meteo.sw_in.copy()
        data_vals["swe"][date] = model.state.snow.swe.copy()

    ds = model.gridded_output.data
    assert_allclose(
        data_vals["sw_in"]["2020-01-17 12:00"],
        ds.sw_in.loc["2020-01-17 12:00", :, :].values,
    )

    assert_allclose(
        data_vals["swe"]["2020-01-19 00:00"],
        ds.swe.loc["2020-01-19 00:00", :, :].values,
    )

    mean_temp = (
        data_vals["temp"]["2020-01-18 00:00"]
        + data_vals["temp"]["2020-01-18 03:00"]
        + data_vals["temp"]["2020-01-18 06:00"]
        + data_vals["temp"]["2020-01-18 09:00"]
        + data_vals["temp"]["2020-01-18 12:00"]
        + data_vals["temp"]["2020-01-18 15:00"]
        + data_vals["temp"]["2020-01-18 18:00"]
        + data_vals["temp"]["2020-01-18 21:00"]
    ) / 8
    assert_allclose(
        mean_temp,
        ds.temp.loc["2020-01-18 21:00", :, :].values,
    )

    precip_sum = (
        data_vals["precip"]["2020-01-18 00:00"]
        + data_vals["precip"]["2020-01-18 03:00"]
        + data_vals["precip"]["2020-01-18 06:00"]
        + data_vals["precip"]["2020-01-18 09:00"]
        + data_vals["precip"]["2020-01-18 12:00"]
        + data_vals["precip"]["2020-01-18 15:00"]
        + data_vals["precip"]["2020-01-18 18:00"]
        + data_vals["precip"]["2020-01-18 21:00"]
    )
    assert_allclose(
        precip_sum,
        ds.precip.loc["2020-01-18 21:00", :, :].values,
    )


def test_data_type(tmp_path):
    config = base_config()
    config.start_date = "2020-01-17"
    config.end_date = "2020-01-17"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {"var": "meteo.temp", "freq": "D", "agg": "mean"},
        {"var": "snow.num_layers", "freq": "D"},
        {"var": "snow.num_layers", "freq": "D", "agg": "sum", "name": "num_layers_sum"},
        {"var": "snow.num_layers", "freq": "D", "agg": "mean", "name": "num_layers_mean"},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    ds = model.gridded_output.data
    assert np.issubdtype(ds.temp.dtype, np.float32)
    assert np.issubdtype(ds.num_layers.dtype, np.integer)
    assert np.issubdtype(ds.num_layers_sum.dtype, np.integer)
    assert np.issubdtype(ds.num_layers_mean.dtype, np.float32)


def test_nothing_to_write(tmp_path):
    config = base_config()
    config.start_date = "2020-01-01"
    config.end_date = "2020-01-01 03:00"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {"var": "snow.swe", "freq": "D", "agg": "sum"},
        {"var": "meteo.temp", "dates": ["2020-01-02 12:00"]},
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    assert model.gridded_output.data is None

    grid_cfg.variables = [
        {"var": "snow.swe", "freq": "D", "agg": "sum"},
        {"var": "meteo.temp", "dates": ["2020-01-01 02:00"]},
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.gridded_output.data
    assert list(ds.data_vars.keys()) == ["temp"]


def test_dask(tmp_path):
    config = base_config()
    config.end_date = "2020-01-16"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = "netcdf"
    grid_cfg.variables = [
        {"var": "meteo.temp", "freq": "D"},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()

    ds_inmem = model.gridded_output._create_dataset(in_memory=True)
    ds_dask = model.gridded_output._create_dataset(in_memory=False)
    assert ds_inmem.temp.chunks is None
    assert ds_dask.temp.chunks is not None


def test_append(tmp_path):
    config = base_config()
    config.start_date = "2020-01-17"
    config.end_date = "2020-01-19"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = "netcdf"
    grid_cfg.variables = [
        {"var": "meteo.rel_hum"},
        {"var": "meteo.temp", "freq": "D", "agg": "mean"},
        {"var": "meteo.precip", "freq": "D", "agg": "sum"},
        {"var": "meteo.sw_in", "dates": ["2020-01-17 12:00"]},
        {"var": "snow.swe", "freq": "D"},
    ]

    grid_cfg.append = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    for date_num in range(len(model.dates)):
        model.run_single()
        if date_num == 0:
            ds = xr.load_dataset(tmp_path / "output_grids.nc")
            assert ds.sizes["time1"] == len(model.dates)
    ds_no_append = xr.load_dataset(tmp_path / "output_grids.nc")

    grid_cfg.append = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    for date_num in range(len(model.dates)):
        model.run_single()
        with xr.open_dataset(tmp_path / "output_grids.nc") as ds:
            assert ds.sizes["time1"] == date_num + 1
    ds_append = xr.load_dataset(tmp_path / "output_grids.nc")

    xr.testing.assert_identical(ds_no_append, ds_append)


@pytest.mark.parametrize("fmt", ["netcdf", "geotiff"])
def test_compress(fmt, tmp_path):
    config = base_config()
    config.start_date = "2020-01-01 00:00"
    config.end_date = "2020-01-01 03:00"
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = fmt
    grid_cfg.variables = [
        {"var": "meteo.temp"},
    ]

    for compress in (False, True):
        grid_cfg.compress = compress
        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()

        if fmt == "netcdf":
            with xr.open_dataset(tmp_path / "output_grids.nc") as ds:
                assert ds["temp"].encoding["zlib"] is compress
        elif fmt == "geotiff":
            with rasterio.open(tmp_path / "temp_2020-01-01T0000.tif") as ds:
                if compress:
                    assert ds.compression is not None
                else:
                    assert ds.compression is None



import pytest

import openamundsen as oa

from .compare import compare_datasets
from .conftest import base_config


@pytest.mark.slow
@pytest.mark.comparison
def test_compare_interpolation_grids():
    bc = base_config()
    bc.start_date = "2019-12-01"
    bc.end_date = "2019-12-01"
    bc.output_data.grids.format = "memory"

    config = bc.copy()
    config.output_data.grids.variables = [{"var": "meteo.temp"}]
    config.meteo.interpolation.temperature.trend_method = "regression"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_temp_regression", model.gridded_output.data)

    config.meteo.interpolation.temperature.trend_method = "fixed"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_temp_fixed", model.gridded_output.data)

    config.meteo.interpolation.temperature.trend_method = "regression"
    config.meteo.interpolation.temperature.regression_params.max_segments = 2
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_temp_piecewise-regression", model.gridded_output.data)

    config = bc.copy()
    config.output_data.grids.variables = [{"var": "meteo.precip"}]
    config.meteo.interpolation.precipitation.trend_method = "regression"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_precip_regression", model.gridded_output.data)

    config.meteo.interpolation.precipitation.trend_method = "adjustment_factor"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_precip_adjustment_factor", model.gridded_output.data)

    config.meteo.interpolation.precipitation.trend_method = "fractional"
    config.meteo.interpolation.precipitation.lapse_rate = [
        0.00048,
        0.00046,
        0.00041,
        0.00033,
        0.00028,
        0.00025,
        0.00024,
        0.00025,
        0.00028,
        0.00033,
        0.00041,
        0.00046,
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_precip_fractional", model.gridded_output.data)

    config = bc.copy()
    config.output_data.grids.variables = [{"var": "meteo.rel_hum"}, {"var": "meteo.dew_point_temp"}]
    config.meteo.interpolation.humidity.trend_method = "regression"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_hum_regression", model.gridded_output.data)

    config.meteo.interpolation.humidity.trend_method = "fixed"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_hum_fixed", model.gridded_output.data)

    config.meteo.interpolation.humidity.trend_method = "regression"
    config.meteo.interpolation.humidity.regression_params.max_segments = 2
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_hum_piecewise-regression", model.gridded_output.data)

    config = bc.copy()
    config.output_data.grids.variables = [{"var": "meteo.wind_speed"}]
    config.meteo.interpolation.wind.trend_method = "regression"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    compare_datasets("interpolation_wind_regression", model.gridded_output.data)

    # config.output_data.grids.variables = [{"var": "meteo.wind_speed"}, {"var": "meteo.wind_dir"}]
    # config.meteo.interpolation.wind.method = "liston"
    # model = oa.OpenAmundsen(config)
    # model.initialize()
    # model.run()
    # compare_datasets("interpolation_wind_liston", model.gridded_output.data)



import numpy as np
from numba import njit, prange


@njit(cache=True, parallel=True)
def shadows(dem, res, sun_vec, num_sweeps=1):
    """
    Calculate terrain shadowing after Corripio (2003).

    Parameters
    ----------
    dem : ndarray
        Terrain elevation (m).

    res : float
        DEM resolution (m).

    sun_vec : ndarray
        Vector describing the position of the sun.

    num_sweeps : int, default 1
        Number of sweeps to perform in each direction. E.g. if num_sweeps = 3,
        sun paths are calculated for all elements in the first 3 rows and
        columns of the DEM in the direction of the sun.
        num_sweeps can be set to -1 to consider all pixels of the DEM.

    Returns
    -------
    illum : ndarray
        Contains 1 for illuminated and 0 for shadowed pixels.

    References
    ----------
    .. [1] Corripio, J. G. (2003). Vectorial algebra algorithms for calculating
       terrain parameters from DEMs and solar radiation modelling in mountainous
       terrain. International Journal of Geographical Information Science, 17(1),
       1–23. https://doi.org/10.1080/13658810210157796
    """
    inv_sun_vec = -sun_vec / np.max(np.abs(sun_vec[:2]))

    normal_sun_vec = np.zeros(3)
    normal_sun_vec[2] = np.sqrt(sun_vec[0] ** 2 + sun_vec[1] ** 2)
    normal_sun_vec[0] = -sun_vec[0] * sun_vec[2] / normal_sun_vec[2]
    normal_sun_vec[1] = -sun_vec[1] * sun_vec[2] / normal_sun_vec[2]

    # Determine origin of scanning lines in the direction of the sun
    if num_sweeps == -1:
        num_sweeps = max(dem.shape)
    pos = np.zeros(dem.shape, dtype=np.uint8)
    if sun_vec[0] < 0 and sun_vec[1] < 0:  # sun is in the Northwest
        pos[:, :num_sweeps] = 1
        pos[:num_sweeps, :] = 1
    elif sun_vec[0] < 0 and sun_vec[1] >= 0:  # sun is in the Southwest
        pos[-num_sweeps:, :] = 1
        pos[:, :num_sweeps] = 1
    elif sun_vec[0] >= 0 and sun_vec[1] < 0:  # sun is in the Northeast
        pos[:num_sweeps, :] = 1
        pos[:, -num_sweeps:] = 1
    elif sun_vec[0] >= 0 and sun_vec[1] >= 0:  # sun is in the Northwest
        pos[-num_sweeps:, :] = 1
        pos[:, -num_sweeps:] = 1
    i_vals, j_vals = np.nonzero(pos)

    illum = np.full(dem.shape, 1, dtype=np.uint8)

    for idx_num in prange(len(i_vals)):
        i = i_vals[idx_num]
        j = j_vals[idx_num]
        n = 0
        max_z_proj = -np.inf

        while True:
            dx = inv_sun_vec[0] * n
            dy = inv_sun_vec[1] * n
            jdx = int(np.round(j + dx))
            idy = int(np.round(i + dy))
            n += 1

            if jdx < 0 or jdx >= dem.shape[1] or idy < 0 or idy >= dem.shape[0]:
                break

            vec_to_orig = np.array(
                [
                    dx * res,
                    dy * res,
                    dem[idy, jdx],
                ]
            )
            z_proj = np.dot(vec_to_orig, normal_sun_vec)

            if z_proj < max_z_proj:
                illum[idy, jdx] = 0
            else:
                max_z_proj = z_proj

    return illum



import numba
import numpy as np


@numba.njit(
    numba.double[:](
        numba.double[:],
        numba.double[:],
        numba.double[:],
        numba.double[:],
        numba.double[:],
        numba.double,
        numba.double,
    ),
    parallel=True,
    cache=True,
)
def _idw(x_points, y_points, z_points, x_targets, y_targets, power, smoothing):
    """
    Interpolate a set or irregularly distributed points using inverse distance
    weighting.
    This function performs the actual interpolation; for the parameters see the
    documentation of `idw` (which is only a wrapper for this function).
    """
    num_points = len(x_points)
    num_targets = len(x_targets)
    data = np.zeros(num_targets)

    for target_num in numba.prange(num_targets):
        x = x_targets[target_num]
        y = y_targets[target_num]
        w = 0.0
        total = 0.0
        dist_is_0 = False

        for k in range(num_points):
            dist = np.sqrt((x - x_points[k]) ** 2 + (y - y_points[k]) ** 2 + smoothing**2)

            if dist == 0.0:
                data[target_num] = z_points[k]
                dist_is_0 = True
                break

            w += 1.0 / dist**power
            total += z_points[k] / dist**power

        if not dist_is_0:
            data[target_num] = total / w

    return data


def idw(x, y, z, x_targets, y_targets, power=2, smoothing=0, ignore_nan=True):
    """
    Interpolate a set or irregularly distributed points using inverse distance
    weighting.

    Parameters
    ----------
    x, y : ndarray
        x and y coordinates of the known points.

    z : ndarray
        Values to be interpolated.

    x_targets, y_targets : ndarray
        x and y coordinates of the interpolation targets.

    power : float, default 2
        Weighting power. The default value of 2 corresponds to the classic
        inverse distance squared weighting.

    smoothing : float, default 0
        Smoothing parameter. Increasing this value will produce smoother
        results, but means that the interpolation is no longer exact.

    ignore_nan : bool, default True
        Ignore NaN values in the interpolation points.

    Returns
    -------
    data : ndarray
        Interpolated values for the target locations.
    """
    x = np.asarray(x, dtype=float)
    y = np.asarray(y, dtype=float)
    z = np.asarray(z, dtype=float)

    if not len(x) == len(y) == len(z):
        raise ValueError("x, y and z must have equal lengths")

    if ignore_nan:
        pos = np.isfinite(x) & np.isfinite(y) & np.isfinite(z)
        x = x[pos]
        y = y[pos]
        z = z[pos]

    # If no input points are available return an all-nan array
    if len(x) == 0:
        return np.full(x_targets.shape, np.nan)

    return _idw(
        x,
        y,
        z,
        np.asarray(x_targets, dtype=float),
        np.asarray(y_targets, dtype=float),
        float(power),
        float(smoothing),
    )



from . import (
    canopy,
    evapotranspiration,
    glacier,
    radiation,
    snow,
    soil,
)

__all__ = [
    "canopy",
    "evapotranspiration",
    "glacier",
    "radiation",
    "snow",
    "soil",
]



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.testing import assert_allclose, assert_array_equal

import openamundsen as oa
from openamundsen import errors
from openamundsen import meteo as oameteo

from .conftest import base_config


def meteo_to_netcdf(ds, out_dir):
    ds = ds.copy(deep=True)
    td = ds.time[1] - ds.time[0]
    timestep = td.values / np.timedelta64(1, "s")

    ds = ds[
        [
            "station_name",
            "lon",
            "lat",
            "alt",
            "time",
            "temp",
            "precip",
            "rel_hum",
            "sw_in",
            "wind_speed",
            "wind_dir",
        ]
    ].rename_vars(
        {
            "temp": "tas",
            "precip": "pr",
            "rel_hum": "hurs",
            "sw_in": "rsds",
            "wind_speed": "wss",
            "wind_dir": "wind_dir",
        }
    )

    ds["pr"] /= timestep
    ds["pr"].attrs = {
        "standard_name": "precipitation_flux",
        "units": "kg m-2 s-1",
    }

    for station_num in range(len(ds.station)):
        ds_cur = ds.isel(station=station_num).copy()
        station_id = str(ds_cur.station.values)
        station_name = str(ds_cur.station_name.values)
        ds_cur.attrs = {
            "Conventions": "CF-1.6",
            "station_name": station_name,
        }
        ds_cur = ds_cur.drop_vars(["station", "station_name"])
        ds_cur.to_netcdf(f"{out_dir}/{station_id}.nc")


def test_formats():
    config = base_config()
    config.start_date = "2015-07-28"
    config.end_date = "2020-12-31"
    config.timestep = "h"

    meta = pd.read_csv(f"{pytest.DATA_DIR}/meteo/rofental/csv/stations.csv", index_col=0)
    meta = meta.sort_index()

    meteo = {}
    for fmt in ("netcdf", "csv"):
        config.input_data.meteo.format = fmt
        config.input_data.meteo.dir = f"{pytest.DATA_DIR}/meteo/rofental/{fmt}"
        model = oa.OpenAmundsen(config)
        model.initialize()
        ds = model.meteo.sortby("station")
        assert ds.time.to_index().equals(model.dates)
        assert_array_equal(meta["name"], ds.station_name)
        assert_array_equal(meta["alt"], ds.alt)
        meteo[fmt] = ds

    xr.testing.assert_allclose(meteo["netcdf"], meteo["csv"])


def test_format_memory(tmp_path):
    config = base_config()
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo

    config.input_data.meteo.format = "memory"
    model = oa.OpenAmundsen(config)
    model.initialize(meteo=oa.forcing.strip_point_dataset(meteo1))
    meteo2 = model.meteo
    assert meteo1.identical(meteo2)

    config.start_date = meteo1.indexes["time"][1]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize(meteo=oa.forcing.strip_point_dataset(meteo1))


def test_format_callback():
    def callback(model):
        model.state.meteo.temp[:] = meteo["temp"][model.date_idx]
        model.state.meteo.precip[:] = meteo["precip"][model.date_idx]
        model.state.meteo.wind_speed[:] = meteo["wind_speed"][model.date_idx]
        model.state.meteo.rel_hum[:] = meteo["rel_hum"][model.date_idx]
        model.state.meteo.cloud_factor[:] = meteo["cloud_factor"][model.date_idx]
        model.state.meteo.cloud_fraction[:] = meteo["cloud_fraction"][model.date_idx]

    config = base_config()
    config.start_date = "2019-12-01"
    config.end_date = "2019-12-15"

    model = oa.OpenAmundsen(config)
    model.initialize()

    for category in model.state.categories:
        for var_name in model.state[category]._meta.keys():
            full_var_name = f"{category}.{var_name}"
            config.output_data.grids.variables.append(
                {
                    "var": full_var_name,
                    "name": full_var_name,
                }
            )

    meteo = {
        "temp": [],
        "precip": [],
        "wind_speed": [],
        "rel_hum": [],
        "cloud_factor": [],
        "cloud_fraction": [],
    }

    model = oa.OpenAmundsen(config)
    model.initialize()
    for _date in model.dates:
        model.run_single()
        meteo["temp"].append(model.state.meteo.temp.copy())
        meteo["precip"].append(model.state.meteo.precip.copy())
        meteo["wind_speed"].append(model.state.meteo.wind_speed.copy())
        meteo["rel_hum"].append(model.state.meteo.rel_hum.copy())
        meteo["cloud_factor"].append(model.state.meteo.cloud_factor.copy())
        meteo["cloud_fraction"].append(model.state.meteo.cloud_fraction.copy())
    data_ref = model.gridded_output.data.copy()

    config.input_data.meteo = {"format": "callback"}
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()
    model.initialize(meteo_callback=callback)
    model.run()
    data_cb = model.gridded_output.data.copy()

    assert data_ref.identical(data_cb)


def test_no_files_found(tmp_path):
    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)

    for fmt in ("netcdf", "csv"):
        config.input_data.meteo.format = fmt
        model = oa.OpenAmundsen(config)
        with pytest.raises(errors.MeteoDataError):
            model.initialize()


def test_missing_csv_metadata_columns(tmp_path):
    config = base_config()
    config.input_data.meteo.format = "csv"
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/csv")
    meta = pd.read_csv(p_orig / "stations.csv", index_col=0)
    meta[["name", "x", "y"]].loc[["bellavista"]].to_csv(tmp_path / "stations.csv")
    (tmp_path / "bellavista.csv").symlink_to(p_orig / "bellavista.csv")

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_inbetween(tmp_path):
    config = base_config()
    config.end_date = "2020-04-30"
    config.timestep = "h"
    config.input_data.meteo.format = "csv"
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/csv")

    meta = pd.read_csv(p_orig / "stations.csv", index_col=0)
    meta.loc[["bellavista"]].to_csv(tmp_path / "stations.csv")

    df = pd.read_csv(p_orig / "bellavista.csv", index_col=0, parse_dates=True)
    df = df.loc[config.start_date : config.end_date]
    df = df.drop(df.index[37:38])
    df.to_csv(tmp_path / "bellavista.csv")

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_start_end():
    config = base_config()
    config.start_date = "1900-01-01"
    config.end_date = "2100-12-31"

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_station_selection(tmp_path):
    config = base_config()
    model = oa.OpenAmundsen(config)
    model.initialize()

    dummy_station = model.meteo.isel(station=0).copy()
    dummy_station["station"] = "dummy"
    dummy_station["station_name"] = "Dummy station"
    dummy_station["lon"] = 11.7
    dummy_station["lat"] = 47.21
    ds = xr.concat([model.meteo, dummy_station], dim="station")
    meteo_to_netcdf(ds, tmp_path)

    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.bounds = "grid"
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ["bellavista", "latschbloder", "proviantdepot"],
    )

    config.input_data.meteo.bounds = "global"
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ["bellavista", "dummy", "latschbloder", "proviantdepot"],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636850,
        5182580,
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ["bellavista"],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636801,
        5182551,
    ]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = "global"
    config.input_data.meteo.exclude = ["bellavista", "latschbloder"]
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ["dummy", "proviantdepot"],
    )

    config.input_data.meteo.bounds = "grid"
    config.input_data.meteo.exclude = ["bellavista", "latschbloder", "proviantdepot"]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = "grid"
    config.input_data.meteo.exclude = []
    config.input_data.meteo.include = ["dummy"]
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ["bellavista", "dummy", "latschbloder", "proviantdepot"],
    )

    config.input_data.meteo.include = []
    config.input_data.meteo.exclude = ["dummy", "dummy2"]
    model = oa.OpenAmundsen(config)
    model.initialize()

    config.input_data.meteo.include = ["dummy", "dummy2"]
    config.input_data.meteo.exclude = []
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_variables(tmp_path):
    ds = xr.load_dataset(f"{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc")
    ds = ds.drop_vars(["tas", "hurs", "rsds", "wss", "wind_dir"])
    ds.to_netcdf(tmp_path / "proviantdepot.nc")

    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()

    assert "temp" in model.meteo
    assert "precip" in model.meteo
    assert "rel_hum" in model.meteo
    assert "sw_in" in model.meteo
    assert "wind_speed" in model.meteo
    assert "wind_dir" not in model.meteo


def test_netcdf_precip_units(tmp_path):
    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)

    ds = xr.load_dataset(f"{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc")
    ds.pr.attrs["units"] = "kg m-2 s-1"
    ds.to_netcdf(tmp_path / "proviantdepot.nc")
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo.copy()

    ds["pr"] *= 3600
    ds.pr.attrs["units"] = "kg m-2"
    ds.to_netcdf(tmp_path / "proviantdepot.nc")
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo2 = model.meteo.copy()

    xr.testing.assert_allclose(meteo1.precip, meteo2.precip)

    ds.pr.attrs["units"] = "mm h-1"
    ds.to_netcdf(tmp_path / "proviantdepot.nc")
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_crs(tmp_path):
    config = base_config()

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds1 = model.meteo[["lon", "lat", "x", "y"]].sortby("station")

    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/csv")
    meta = pd.read_csv(p_orig / "stations.csv", index_col=0)
    x_new, y_new = oa.util.transform_coords(meta.x, meta.y, "epsg:32632", "epsg:3416")
    meta.x = x_new
    meta.y = y_new
    meta.to_csv(tmp_path / "stations.csv")

    for station_id in meta.index:
        (tmp_path / f"{station_id}.csv").symlink_to(p_orig / f"{station_id}.csv")

    config.input_data.meteo.format = "csv"
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.crs = "epsg:3416"

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo[["lon", "lat", "x", "y"]].sortby("station")

    xr.testing.assert_allclose(ds1, ds2)


def test_grid_cell_assignment(tmp_path):
    config = base_config()
    config.start_date = "2020-01-01"
    config.end_date = "2020-01-01"

    model = oa.OpenAmundsen(config)
    model.initialize()
    grid_xs = model.grid.xs
    grid_ys = model.grid.ys

    config.input_data.meteo.format = "csv"
    config.input_data.meteo.dir = str(tmp_path)

    station_coords_colrows = [  # (x, y, col, row)
        (grid_xs[0], grid_ys[0], 0, 0),
        (grid_xs[0] + 1, grid_ys[0] - 1, 0, 0),
        (grid_xs[1] + 1, grid_ys[0] - 1, 1, 0),
        (grid_xs[1] + 1, grid_ys[3] - 1, 1, 3),
        (grid_xs[-1] + 1, grid_ys[-1] - 1, len(grid_xs) - 1, len(grid_ys) - 1),
        (grid_xs[0], grid_ys[-1], 0, len(grid_ys) - 1),
    ]

    meta = pd.DataFrame(
        index=[f"dummy{i}" for i in range(len(station_coords_colrows))],
        data={
            "name": None,
            "x": [s[0] for s in station_coords_colrows],
            "y": [s[1] for s in station_coords_colrows],
            "alt": 0.0,
        },
    )
    meta.to_csv(tmp_path / "stations.csv")

    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/csv")
    for station_id in meta.index:
        (tmp_path / f"{station_id}.csv").symlink_to(p_orig / "proviantdepot.csv")

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds = model.meteo

    for i, s in enumerate(station_coords_colrows):
        ds_s = ds.sel(station=f"dummy{i}")
        expected_col = s[2]
        expected_row = s[3]
        assert int(ds_s.col) == expected_col
        assert int(ds_s.row) == expected_row


@pytest.mark.parametrize("fmt", ["netcdf", "csv"])
def test_slice_and_resample(fmt, tmp_path):
    def meteo_to_df(model):
        return (model.meteo.sel(station="bellavista").to_dataframe()[params]).astype(float)

    def matches_start_and_end_date(model):
        idx = model.meteo.indexes["time"]
        return idx[0] == model.config.start_date and idx[-1] == model.config.end_date

    params = ["temp", "precip", "rel_hum", "sw_in", "wind_speed"]

    agg_funcs_inst = {p: lambda s: s.iloc[-1] for p in params}
    agg_funcs_inst["precip"] = pd.Series.sum
    agg_funcs_res = dict.fromkeys(params, pd.Series.mean)
    agg_funcs_res["precip"] = pd.Series.sum

    config = base_config()
    config.start_date = "2015-11-19"
    config.end_date = "2015-11-30"
    config.timestep = "h"
    config.input_data.meteo.format = fmt
    config.input_data.meteo.dir = f"{pytest.DATA_DIR}/meteo/rofental/{fmt}"
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_h = meteo_to_df(model)

    config.start_date = "2015-11-20"
    config.end_date = "2015-11-30"
    config.timestep = "3h"
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-30 13:00":"2015-11-30 15:00"].agg(agg_funcs_inst, skipna=False),
        df_res.loc["2015-11-30 15:00"],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc["2015-11-26 15:00"].precip)
    assert_allclose(
        df_h.loc["2015-11-19 22:00":"2015-11-30 21:00"].precip.sum(),
        df_res.precip.sum(),
    )

    config.start_date = "2015-11-21"
    config.end_date = "2015-11-30"
    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-20 22:00":"2015-11-21 00:00"].agg(agg_funcs_res, skipna=False),
        df_res.loc["2015-11-21 00:00"],
        check_exact=False,
        check_names=False,
    )
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-30 13:00":"2015-11-30 15:00"].agg(agg_funcs_res, skipna=False),
        df_res.loc["2015-11-30 15:00"],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc["2015-11-26 15:00"].precip)

    config.start_date = "2015-11-25"
    config.end_date = "2015-11-29"
    config.timestep = "D"
    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-25 01:00":"2015-11-26 00:00"].agg(agg_funcs_res, skipna=False),
        df_res.loc["2015-11-25"],
        check_exact=False,
        check_names=False,
    )
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-29 01:00":"2015-11-30 00:00"].agg(agg_funcs_res, skipna=False),
        df_res.loc["2015-11-29"],
        check_exact=False,
        check_names=False,
    )

    config.start_date = "2015-11-01"
    config.end_date = "2015-11-09"
    config.timestep = "2D"
    wf = config.output_data.timeseries.write_freq
    config.output_data.timeseries.write_freq = "10D"
    model = oa.OpenAmundsen(config)
    config.output_data.timeseries.write_freq = wf
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = "2015-11-01 01:00"
    config.end_date = "2015-11-30 22:00"
    config.timestep = "3h"
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-30 14:00":"2015-11-30 16:00"].agg(agg_funcs_inst, skipna=False),
        df_res.loc["2015-11-30 16:00"],
        check_exact=False,
        check_names=False,
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc["2015-11-30 14:00":"2015-11-30 16:00"].agg(agg_funcs_res, skipna=False),
        df_res.loc["2015-11-30 16:00"],
        check_exact=False,
        check_names=False,
    )

    config.start_date = "2015-11-01 01:37"
    config.end_date = "2015-11-30 22:37"
    config.timestep = "3h"
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = "2015-11-29"
    config.end_date = "2015-11-30"
    config.timestep = "10min"
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    # Wind direction
    config.start_date = "2015-12-20"
    config.end_date = "2015-12-30"
    config.timestep = "h"
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_h = model.meteo.sel(station="bellavista").to_dataframe()[["wind_speed", "wind_dir"]]

    config.timestep = "3h"
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = model.meteo.sel(station="bellavista").to_dataframe()[["wind_speed", "wind_dir"]]
    assert np.allclose(
        df_h["wind_dir"].loc["2015-12-20 15:00"],
        df_res["wind_dir"].loc["2015-12-20 15:00"],
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = model.meteo.sel(station="bellavista").to_dataframe()[["wind_speed", "wind_dir"]]
    wind_us, wind_vs = oameteo.wind_to_uv(
        df_h["wind_speed"].loc["2015-12-20 13:00":"2015-12-20 15:00"],
        df_h["wind_dir"].loc["2015-12-20 13:00":"2015-12-20 15:00"],
    )
    wind_u_mean = wind_us.mean()
    wind_v_mean = wind_vs.mean()
    _, wind_dir = oameteo.wind_from_uv(wind_u_mean, wind_v_mean)
    assert np.allclose(
        df_res["wind_dir"].loc["2015-12-20 15:00"],
        wind_dir,
    )


def test_aggregation_when_using_csv_input_and_start_date_before_station_start_date(tmp_path):
    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/csv")
    meta = pd.read_csv(p_orig / "stations.csv", index_col=0)
    meta.loc[["bellavista", "latschbloder"]].to_csv(tmp_path / "stations.csv")
    (tmp_path / "latschbloder.csv").symlink_to(p_orig / "latschbloder.csv")
    df = pd.read_csv(p_orig / "bellavista.csv", index_col=0, parse_dates=True)
    df.loc["2015-12-22":].to_csv(tmp_path / "bellavista.csv")

    config = base_config()
    config.start_date = "2015-12-20"
    config.end_date = "2015-12-30"
    config.timestep = "D"
    config.input_data.meteo.format = "csv"
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.aggregate_when_downsampling = True

    model = oa.OpenAmundsen(config)
    model.initialize()


def test_aggregation_to_daily_data_with_subdaily_input():
    config = base_config()
    config.start_date = "2015-12-20"
    config.end_date = "2015-12-30"
    config.timestep = "D"

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo

    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo2 = model.meteo

    xr.testing.assert_equal(meteo1, meteo2)


def test_non_hourly_input(tmp_path):
    config = base_config()
    config.start_date = "2015-11-20"
    config.end_date = "2015-11-30"
    config.timestep = "3h"

    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo_to_netcdf(model.meteo, tmp_path)
    ds1 = model.meteo

    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo

    xr.testing.assert_allclose(ds1, ds2)


@pytest.mark.parametrize("aggregate", [False, True])
def test_resample_with_non_matching_start_date(aggregate, tmp_path):
    ds = xr.load_dataset(f"{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc")
    ds = ds.sel(time=slice("2020-11-03 02:00", None))
    ds.to_netcdf(tmp_path / "proviantdepot.nc")

    p_orig = Path(f"{pytest.DATA_DIR}/meteo/rofental/netcdf")
    for station_id in ("bellavista", "latschbloder"):
        (tmp_path / f"{station_id}.nc").symlink_to(p_orig / f"{station_id}.nc")

    config = base_config()
    config.start_date = "2020-11-01"
    config.end_date = "2020-11-30"
    config.timestep = "3h"
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.aggregate_when_downsampling = aggregate

    model = oa.OpenAmundsen(config)
    model.initialize()
    assert model.meteo.indexes["time"].equals(model.dates)


def test_filters(tmp_path):
    config = base_config()
    config.start_date = "2015-07-28"
    config.end_date = "2020-12-31"
    config.timestep = "h"
    default_filters = config.input_data.meteo.filters

    config.input_data.meteo.filters = []
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds_unfiltered = model.meteo

    config.input_data.meteo.filters = default_filters
    model = oa.OpenAmundsen(config)
    model.initialize()
    xr.testing.assert_identical(ds_unfiltered, model.meteo)

    ds_unfiltered["temp"].loc["bellavista", "2015-08-01 12:00":"2015-08-02 23:00"] = 15
    ds_unfiltered["rel_hum"].loc["bellavista", "2015-08-03 12:00"] = 100
    ds_unfiltered["rel_hum"].loc["bellavista", "2015-08-03 13:00"] = 100.1
    ds_unfiltered["precip"].loc["bellavista", "2015-08-03 12:00"] = -10
    ds_unfiltered["precip"].loc["bellavista", "2015-08-03 13:00"] = 10000

    meteo_to_netcdf(ds_unfiltered, tmp_path)
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds_filtered = model.meteo
    assert all(
        np.isnan(ds_filtered["temp"].loc["bellavista", "2015-08-01 12:00":"2015-08-02 23:00"])
    )
    assert ds_filtered["rel_hum"].loc["bellavista", "2015-08-03 12:00"] == 100
    assert np.isnan(ds_filtered["rel_hum"].loc["bellavista", "2015-08-03 13:00"])
    assert np.isnan(ds_filtered["precip"].loc["bellavista", "2015-08-03 12:00"])
    assert ds_filtered["precip"].loc["bellavista", "2015-08-03 13:00"] == 10000
    xr.testing.assert_identical(
        ds_unfiltered.sel(time=slice("2015-08-04 00:00", None)),
        ds_filtered.sel(time=slice("2015-08-04 00:00", None)),
    )


def test_one_station_ends_less_than_one_day_before_start_date(tmp_path):
    # Test a fix for an issue introduced in the improved aggregation handling from #106 which would
    # occur when the data for one station ends less than one day before the simulation start date.
    config = base_config()
    config.start_date = "2015-11-20"
    config.end_date = "2015-11-30"
    config.timestep = "h"

    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo_to_netcdf(model.meteo, tmp_path)

    ds_bv = xr.load_dataset(tmp_path / "bellavista.nc")
    ds_bv.sel(time=slice(None, "2015-11-26 11:00")).to_netcdf(tmp_path / "bellavista.nc")

    config.start_date = "2015-11-27"
    config.timestep = "3h"
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()



DAYS_PER_YEAR = 365.25
HOURS_PER_DAY = 24
MINUTES_PER_HOUR = 60
SECONDS_PER_HOUR = 3600

STANDARD_TIMEZONE_WIDTH = 15
SUN_DEGREES_PER_HOUR = 15  # 360/24
MINUTES_PER_DEGREE_OF_EARTH_ROTATION = 4  # earth rotates 1 degree every 4 minutes
SOLAR_CONSTANT = 1366.1  # W m-2

STANDARD_ATMOSPHERE = 101_325  # Pa
ATMOSPHERIC_LAPSE_RATE = 6.5e-3  # K m-1
GRAVITATIONAL_ACCELERATION = 9.80665  # m s-2

T0 = 273.15  # K

MOLAR_MASS_DRY_AIR = 0.02896968  # kg mol-1
MOLAR_MASS_WATER = 0.0180153  # kg mol-1
UNIVERSAL_GAS_CONSTANT = 8.314462618  # J mol-1 K-1
GAS_CONSTANT_DRY_AIR = 287.058  # J kg-1 K-1
SPEC_GAS_CONSTANT_WATER_VAPOR = 461.52  # J kg-1 K-1
STANDARD_SEA_LEVEL_TEMPERATURE = 15 + T0  # K

SPEC_HEAT_CAP_DRY_AIR = 1004.68506  # specific heat capacity of dry air (J kg-1 K-1)
SPEC_HEAT_CAP_ICE = 2100.0  # specific heat capacity of ice (J kg-1 K-1)
SPEC_HEAT_CAP_WATER = 4180.0  # specific heat capacity of water (J kg-1 K-1)

LATENT_HEAT_OF_VAPORIZATION = 2.501e6  # J kg-1 (= latent heat of condensation)
LATENT_HEAT_OF_FUSION = 0.334e6  # J kg-1
LATENT_HEAT_OF_SUBLIMATION = LATENT_HEAT_OF_VAPORIZATION + LATENT_HEAT_OF_FUSION  # J kg-1

STEFAN_BOLTZMANN = 5.670374419e-8  # Stefan-Boltzmann constant (W m-2 K-4)
VON_KARMAN = 0.40  # von Karman constant

WATER_DENSITY = 1000.0  # density of water (kg m-3)
ICE_DENSITY = 917.0  # density of ice (kg m-3)

# Soil constants
VOL_HEAT_CAP_SAND = 2.128e6  # volumetric heat capacity of sand (J m-3 K-1)
VOL_HEAT_CAP_CLAY = 2.385e6  # volumetric heat capacity of clay (J m-3 K-1)

# Thermal conductivities
THERM_COND_AIR = 0.025  # thermal conductivity of air (W m-1 K-1)
THERM_COND_CLAY = 1.16  # thermal conductivity of clay (W m-1 K-1)
THERM_COND_ICE = 2.24  # thermal conducivity of ice (W m-1 K-1)
THERM_COND_SAND = 1.57  # thermal conductivity of sand (W m-1 K-1)
THERM_COND_WATER = 0.56  # thermal conductivity of water (W m-1 K-1)

CRS_WGS84 = "epsg:4326"

METEO_VAR_METADATA = {
    "temp": {
        "standard_name": "air_temperature",
        "units": "K",
    },
    "precip": {
        "standard_name": "precipitation_amount",
        "units": "kg m-2",
    },
    "rel_hum": {
        "standard_name": "relative_humidity",
        "units": "%",
    },
    "sw_in": {
        "standard_name": "surface_downwelling_shortwave_flux_in_air",
        "units": "W m-2",
    },
    "wind_speed": {
        "standard_name": "wind_speed",
        "units": "m s-1",
    },
    "wind_speed_gust": {
        "standard_name": "wind_speed_of_gust",
        "units": "m s-1",
    },
    "wind_dir": {
        "standard_name": "wind_from_direction",
        "units": "degree",
    },
    "cloud_fraction": {
        "standard_name": "cloud_area_fraction",
        "units": "%",
    },
}
MINIMUM_REQUIRED_METEO_VARS = ["temp", "precip", "rel_hum", "sw_in", "wind_speed"]

NETCDF_VAR_MAPPINGS = {
    "tas": "temp",
    "pr": "precip",
    "hurs": "rel_hum",
    "rsds": "sw_in",
    "wss": "wind_speed",
    "wind_dir": "wind_dir",
    "wind_speed_gust": "wind_speed_gust",
}

# Mappings of internal variable names to interpolation config keys (e.g.
# config['meteo']['interpolation']['temperature'])
INTERPOLATION_CONFIG_PARAM_MAPPINGS = {
    "temp": "temperature",
    "precip": "precipitation",
    "rel_hum": "humidity",
}

ALLOWED_METEO_VAR_RANGES = {
    "temp": (0, None),
    "precip": (0, None),
    "rel_hum": (0, 100),
    "sw_in": (0, None),
    "wind_speed": (0.1, None),
    "wind_speed_gust": (0.1, None),
    "cloud_fraction": (0, 100),
}



from dataclasses import dataclass

import netCDF4
import numpy as np
import pandas as pd
import pyproj
import xarray as xr
from loguru import logger

from openamundsen import constants, errors, fileio, util

try:
    import dask.array

    _DASK_AVAILABLE = True
except ImportError:
    _DASK_AVAILABLE = False


_ALLOWED_OFFSETS = [
    pd.tseries.offsets.YearEnd,
    pd.tseries.offsets.YearBegin,
    pd.tseries.offsets.MonthEnd,
    pd.tseries.offsets.MonthBegin,
    pd.tseries.offsets.Day,
    pd.tseries.offsets.Hour,
    pd.tseries.offsets.Minute,
]


@dataclass
class OutputField:
    """
    Class for defining an output field, i.e., a state variable that should be
    written at specified dates.

    Parameters
    ----------
    var : str
        Name of the state variable (e.g. "meteo.temp").

    output_name : str
        Output name.

    agg : str, optional
        Aggregation function. Can be either None (if instantaneous values
        should be written), "sum" or "mean".

    write_dates : pd.DatetimeIndex
        Dates at which the field should be written.

    data : np.array, optional
        Current state of the aggregated values (only used if `agg` is not None).

    num_aggregations : int, default 0
        Current number of aggregations (required for calculating a running mean).
    """

    var: str
    output_name: str
    agg: str
    write_dates: pd.DatetimeIndex
    data: np.array = None
    num_aggregations: int = 0


def _field_key(field):
    return (tuple(field.write_dates), field.agg is None)


class GriddedOutputManager:
    """
    Class for managing and storing gridded output data which should be written at
    specified dates.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.
    """

    def __init__(self, model):
        config = model.config.output_data.grids
        fields = []

        for field_cfg in config.variables:
            try:
                output_name = field_cfg["name"]
            except KeyError:
                output_name = None

            try:
                freq = field_cfg["freq"]
            except KeyError:
                freq = model.dates.freqstr

            try:
                agg = field_cfg["agg"]
            except KeyError:
                agg = None

            if "dates" in field_cfg:
                write_dates = pd.to_datetime(field_cfg["dates"])
            else:
                write_dates = _freq_write_dates(model.dates, freq, agg is not None)

            write_dates = write_dates[
                (write_dates >= model.dates[0]) & (write_dates <= model.dates[-1])
            ]
            if len(write_dates) == 0:
                logger.debug(
                    f"Discarding grid output variable {field_cfg['var']} (nothing to be written)"
                )
                continue

            if output_name is None:
                output_name = field_cfg.var.split(".")[-1]

            if output_name in [f.output_name for f in fields]:
                raise errors.ConfigurationError(f"Duplicate grid output name: {output_name}")

            fields.append(
                OutputField(
                    var=field_cfg["var"],
                    output_name=output_name,
                    agg=agg,
                    write_dates=write_dates,
                )
            )

        self.model = model
        self.fields = fields
        self.format = config.format
        self.nc_file_created = False
        self.data = None

    def update(self):
        """
        Update the output fields for the current time step, i.e., update the
        aggregated fields (if aggregation functions are used) and write the
        variables to file at the specified dates.
        """
        # If there is nothing to be written, return right away
        if len(self.fields) == 0:
            return

        logger.debug("Updating field outputs")

        date = self.model.date
        roi = self.model.grid.roi

        if self.format == "netcdf":
            nc_file = self.model.config.results_dir / "output_grids.nc"
            append = self.model.config.output_data.grids.append

            if not self.nc_file_created:
                ds = self._create_dataset(in_memory=(not _DASK_AVAILABLE))

                if append:
                    # Convert time values to encoded times (e.g. X hours since YYYY-MM-DD)
                    self._time_dims = [d for d in list(ds.dims) if d.startswith("time")]
                    self._encoded_times = {}
                    for coord_name in list(ds.coords):
                        if not coord_name.startswith("time"):
                            continue

                        encoded_times, _, _ = xr.coding.times.encode_cf_datetime(
                            ds.coords[coord_name],
                            units=ds[self._time_dims[0]].encoding["units"],
                            calendar=ds[self._time_dims[0]].encoding["calendar"],
                        )
                        # (taking units and calendar from the first time variable works because all
                        # time variables have the same encoding)
                        self._encoded_times[coord_name] = encoded_times

                    ds = ds.drop_sel({v: ds[v] for v in self._time_dims})
                    ds.encoding["unlimited_dims"] = self._time_dims

                ds.to_netcdf(nc_file)
                self.nc_file_created = True

            ds = netCDF4.Dataset(nc_file, "r+")
        elif self.format == "memory":
            if self.data is None:
                self.data = self._create_dataset(in_memory=True)

        # Loop through all fields, update aggregations where necessary and write files at the
        # specified dates
        for field in self.fields:
            dtype, fill_value = self._field_dtype_and_fill_value(field)

            # Keep float values in float64 for calculating sums and averages, only cast to float32
            # when writing
            if np.issubdtype(dtype, np.floating):
                dtype = np.float64

            if field.agg is not None:
                if field.data is None:
                    meta = self.model.state.meta(field.var)

                    if meta.dim3 == 0:
                        arr = np.full(self.model.grid.shape, fill_value, dtype=dtype)
                        arr[roi] = 0
                    else:
                        arr = np.full((meta.dim3, *self.model.grid.shape), fill_value, dtype=dtype)
                        arr[:, roi] = 0

                    field.data = arr

                data_cur = self.model.state[field.var]

                if field.agg == "sum":
                    if field.data.ndim == 2:
                        field.data[roi] += data_cur[roi]
                    else:
                        field.data[:, roi] += data_cur[:, roi]
                elif field.agg == "mean":
                    if field.data.ndim == 2:
                        field.data[roi] += (data_cur[roi] - field.data[roi]) / (
                            field.num_aggregations + 1
                        )
                    else:
                        field.data[:, roi] += (data_cur[:, roi] - field.data[:, roi]) / (
                            field.num_aggregations + 1
                        )

                field.num_aggregations += 1

            if date in field.write_dates:
                date_idx = np.flatnonzero(field.write_dates == date)[0]

                if field.agg is None:
                    data = self.model.state[field.var]
                else:
                    data = field.data

                if self.format == "netcdf":
                    ds[field.output_name][date_idx, :, :] = data

                    if append:
                        field_time_dim = ds[field.output_name].dimensions[0]
                        ds[field_time_dim][date_idx] = self._encoded_times[field_time_dim][date_idx]

                        bounds_var = f"{field_time_dim}_bounds"
                        if bounds_var in self._encoded_times:
                            ds[bounds_var][date_idx, :] = self._encoded_times[bounds_var][date_idx]
                elif self.format in ("ascii", "geotiff"):
                    if self.format == "ascii":
                        ext = "asc"
                        rio_meta = {"driver": "AAIGrid"}
                        # (do not add CRS information when using AAIGrid output to avoid writing
                        # .prj files)
                    elif self.format == "geotiff":
                        ext = "tif"
                        rio_meta = {
                            "driver": "GTiff",
                            "crs": self.model.grid.crs,
                        }

                        if self.model.config.output_data.grids.compress:
                            rio_meta["compress"] = "lzw"

                    if field.agg is None:
                        date_str = f"{date:%Y-%m-%dT%H%M}"
                    else:
                        # Find the start date of the current output interval for the output file
                        # name
                        if date_idx == 0:
                            start_date = self.model.dates[0]
                        else:
                            start_date = field.write_dates[date_idx - 1] + pd.Timedelta(
                                seconds=self.model.timestep
                            )

                        date_str = f"{start_date:%Y-%m-%dT%H%M}_{date:%Y-%m-%dT%H%M}"

                    if data.ndim == 2:
                        filename = (
                            self.model.config.results_dir / f"{field.output_name}_{date_str}.{ext}"
                        )
                        logger.debug(f"Writing field {field.var} to {filename}")
                        fileio.write_raster_file(
                            filename,
                            data,
                            self.model.grid.transform,
                            **rio_meta,
                        )
                    else:
                        # For 3-dimensional variables, write each layer as a separate file
                        for layer_num in range(data.shape[0]):
                            filename = (
                                self.model.config.results_dir
                                / f"{field.output_name}_{layer_num}_{date_str}.{ext}"
                            )
                            logger.debug(
                                f"Writing field {field.var} (layer {layer_num}) to {filename}"
                            )
                            fileio.write_raster_file(
                                filename,
                                data[layer_num, :, :],
                                self.model.grid.transform,
                                **rio_meta,
                            )
                elif self.format == "memory":
                    self.data[field.output_name].values[date_idx, :, :] = data
                else:
                    raise NotImplementedError

                field.data = None
                field.num_aggregations = 0

        if self.format == "netcdf":
            ds.close()

    def _create_dataset(self, in_memory=False):
        """
        Create a CF-compliant Dataset covering the specified output variables
        and dates.

        Parameters
        ----------
        in_memory : bool, default False
            If True use in-memory arrays for creating the dataset, if False use
            Dask arrays.

        Returns
        -------
        ds : xr.Dataset
        """
        # Define names of time variables - if there is only one time variable simply name it "time",
        # otherwise they are named "time1", "time2", ...
        time_var_names = {}
        num_time_vars = 0
        for field in self.fields:
            key = _field_key(field)
            if key not in time_var_names:
                num_time_vars += 1
                time_var_names[key] = f"time{num_time_vars}"

        if num_time_vars == 1:
            key = next(iter(time_var_names))
            time_var_names[key] = "time"

        times = {}  # dict for storing times and boundaries (for aggregated variables) of the time variables # noqa: E501
        field_time_vars = []  # contains for each field the name of the respective NetCDF time variable # noqa: E501

        for field in self.fields:
            key = _field_key(field)
            time_var_name = time_var_names[key]
            time_vals = field.write_dates.values

            if field.agg is None:
                time_vals = field.write_dates
                time_bounds = None
            else:
                time_bounds = np.repeat(time_vals[:, np.newaxis], 2, axis=1).copy()
                time_bounds[1:, 0] = time_bounds[:-1, 1]
                time_bounds[0, 0] = self.model.dates[0]

            field_time_vars.append(time_var_name)

            if time_var_name not in times:
                times[time_var_name] = (time_vals, time_bounds)

        x_coords = self.model.grid.X[0, :]
        y_coords = self.model.grid.Y[:, 0]

        # Define coordinate variables
        coords = {}
        for time_var, (time_vals, time_bounds) in times.items():
            time_attrs = {}

            if time_bounds is not None:
                bound_var_name = f"{time_var}_bounds"
                time_attrs["bounds"] = bound_var_name
                coords[bound_var_name] = (
                    [time_var, "nbnd"],
                    time_bounds,
                    {
                        "long_name": "time interval endpoints",
                    },
                )

            coords[time_var] = (
                time_var,
                time_vals,
                time_attrs,
            )

        coords["x"] = (
            ["x"],
            x_coords,
            {
                "standard_name": "projection_x_coordinate",
                "long_name": "x coordinate of projection",
                "units": "m",
            },
        )
        coords["y"] = (
            ["y"],
            y_coords,
            {
                "standard_name": "projection_y_coordinate",
                "long_name": "y coordinate of projection",
                "units": "m",
            },
        )
        coords["crs"] = (
            [],
            np.array(0),
            pyproj.crs.CRS(self.model.grid.crs).to_cf(),
        )

        if in_memory:
            full = np.full
        else:
            full = dask.array.full

        if self.format != "memory" and not _DASK_AVAILABLE:
            dataset_size = 0
            for field in self.fields:
                meta = self.model.state.meta(field.var)
                dtype, _ = self._field_dtype_and_fill_value(field)
                dataset_size += (
                    np.dtype(dtype).itemsize
                    * len(field.write_dates)
                    * max(meta.dim3, 1)
                    * len(y_coords)
                    * len(x_coords)
                )

            dataset_size_gib = dataset_size / 1024**3
            if dataset_size_gib > 4:
                logger.warning(
                    f"Creating large in-memory dataset ({dataset_size_gib:.1f} GiB) for gridded "
                    "outputs. If you run into memory issues consider installing Dask "
                    "(pip install dask) to enable memory-efficient dataset creation."
                )

        # Define data variables
        data = {}
        three_dim_coords = {}
        for field, field_time_var in zip(self.fields, field_time_vars):
            meta = self.model.state.meta(field.var)
            attrs = {}

            for attr in ("standard_name", "long_name", "units"):
                attr_val = getattr(meta, attr)
                if attr_val is not None:
                    attrs[attr] = attr_val

            attrs["grid_mapping"] = "crs"

            dtype, fill_value = self._field_dtype_and_fill_value(field)
            attrs["_FillValue"] = fill_value

            if meta.dim3 == 0:  # 2-dimensional variable
                data[field.output_name] = (
                    [field_time_var, "y", "x"],
                    full(
                        (len(field.write_dates), len(y_coords), len(x_coords)),
                        fill_value,
                        dtype=dtype,
                    ),
                    attrs,
                )
            else:  # 3-dimensional variable
                category = self.model.state.parse(field.var)[0]
                coord_name = f"{category}_layer"

                if category in three_dim_coords:
                    if three_dim_coords[coord_name] != meta.dim3:
                        # We assume that all 3-dimensional variables within a category have the
                        # same shape (e.g. "soil.temp" must have the same shape as
                        # "soil.therm_cond"); varying numbers of layers within a category are not
                        # supported
                        raise ValueError("Inconsistent length of third variable dimension")
                else:
                    three_dim_coords[coord_name] = meta.dim3

                data[field.output_name] = (
                    [field_time_var, coord_name, "y", "x"],
                    full(
                        (len(field.write_dates), meta.dim3, len(y_coords), len(x_coords)),
                        fill_value,
                        dtype=dtype,
                    ),
                    attrs,
                )

        # Add 3-dimensional coordinates
        for coord_name, coord_len in three_dim_coords.items():
            coords[coord_name] = ([coord_name], np.arange(coord_len))

        ds = xr.Dataset(data, coords=coords)
        ds.attrs["Conventions"] = "CF-1.7"

        _, datetime_units, calendar = xr.coding.times.encode_cf_datetime(self.model.dates)
        for time_var in times:
            ds[time_var].attrs["standard_name"] = "time"

            # Set time units manually because otherwise the units of the time and the time bounds
            # variables might be different which is not recommended by CF standards
            ds[time_var].encoding["units"] = datetime_units
            ds[time_var].encoding["calendar"] = calendar

            # Store time variables as doubles for CF compliance
            ds[time_var].encoding["dtype"] = np.float64
            if f"{time_var}_bounds" in ds:
                ds[f"{time_var}_bounds"].encoding["dtype"] = np.float64

        if self.model.config.output_data.grids.compress:
            for data_var in ds.data_vars:
                ds[data_var].encoding["zlib"] = True
                ds[data_var].encoding["complevel"] = 1

        return ds

    def _field_dtype_and_fill_value(self, field: OutputField):
        """
        Return the dtype and fill value for a given output field.
        Float-like variables are written as float32, integer variables are written as int32 or
        float32 (the latter if agg == 'mean').

        Parameters
        ----------
        field : OutputField
            Field variable.

        Returns
        -------
        A (dtype, fill_value) tuple.
        """
        if (
            np.issubdtype(self.model.state.meta(field.var).dtype, np.integer)
            and field.agg != "mean"
        ):
            dtype = np.int32
            fill_value = np.iinfo(dtype).min
        else:
            dtype = np.float32
            fill_value = np.nan

        return (dtype, fill_value)


def _freq_write_dates(dates, out_freq, agg):
    """
    Calculate output dates for gridded outputs when a frequency string is set.

    For non-aggregated fields the write dates are assigned to the start of the
    respective intervals for non-anchored and begin-anchored offsets (e.g. 'D',
    'MS', 'YS'), and to the end of the intervals for end-anchored offsets (e.g.
    'ME', 'YE'). For aggregated fields, the write dates are always assigned to the
    end of the intervals.

    Parameters
    ----------
    dates : pd.DatetimeIndex
        Simulation dates.

    out_freq : str
        Output frequency as a pandas offset string (e.g. '3h', 'ME').

    agg : boolean
        Prepare write dates for aggregated outputs (if True) or for
        instantaneous values.

    Returns
    -------
    write_dates : pd.DatetimeIndex

    Examples
    --------
    >>> dates = pd.date_range(
    ...     start='2021-01-01 00:00',
    ...     end='2021-12-31 23:00',
    ...     freq='h',
    ... )
    ... _freq_write_dates(dates, 'YE', False)
    DatetimeIndex(['2021-12-31 23:00:00'], dtype='datetime64[ns]', freq=None)

    >>> _freq_write_dates(dates, 'YS', False)
    DatetimeIndex(['2021-01-01'], dtype='datetime64[ns]', freq='AS-JAN')

    >>> _freq_write_dates(dates, 'D', False)
    DatetimeIndex(['2021-01-01', '2021-01-02', '2021-01-03', '2021-01-04',
                   '2021-01-05', '2021-01-06', '2021-01-07', '2021-01-08',
                   '2021-01-09', '2021-01-10',
                   ...
                   '2021-12-22', '2021-12-23', '2021-12-24', '2021-12-25',
                   '2021-12-26', '2021-12-27', '2021-12-28', '2021-12-29',
                   '2021-12-30', '2021-12-31'],
                  dtype='datetime64[ns]', length=365, freq='D')

    >>> _freq_write_dates(dates, 'D', True)
    DatetimeIndex(['2021-01-01 23:00:00', '2021-01-02 23:00:00',
                   '2021-01-03 23:00:00', '2021-01-04 23:00:00',
                   '2021-01-05 23:00:00', '2021-01-06 23:00:00',
                   '2021-01-07 23:00:00', '2021-01-08 23:00:00',
                   '2021-01-09 23:00:00', '2021-01-10 23:00:00',
                   ...
                   '2021-12-22 23:00:00', '2021-12-23 23:00:00',
                   '2021-12-24 23:00:00', '2021-12-25 23:00:00',
                   '2021-12-26 23:00:00', '2021-12-27 23:00:00',
                   '2021-12-28 23:00:00', '2021-12-29 23:00:00',
                   '2021-12-30 23:00:00', '2021-12-31 23:00:00'],
                  dtype='datetime64[ns]', length=365, freq='D')
    """
    model_freq = dates.freqstr
    model_freq_td = util.offset_to_timedelta(model_freq)

    try:
        out_offset = util.to_offset(out_freq)
        if not any(isinstance(out_offset, o) for o in _ALLOWED_OFFSETS):
            raise ValueError  # noqa: TRY301
    except ValueError as err:
        allowed_offsets_str = ", ".join([o().__class__.__name__ for o in _ALLOWED_OFFSETS])
        raise errors.ConfigurationError(
            f"Unsupported output frequency: {out_freq}. Supported offsets: {allowed_offsets_str}"
        ) from err

    if not _is_anchored(out_offset):
        # For non-anchored offsets (e.g., '3h', 'D'), the output frequency must be a multiple of
        # (and not smaller than) the model timestep
        out_freq_td = util.offset_to_timedelta(out_freq)

        if out_freq_td < model_freq_td:
            raise ValueError("Output frequency must not be smaller than the model timestep")
        elif not (out_freq_td.total_seconds() / model_freq_td.total_seconds()).is_integer():
            raise ValueError("Output frequency must be a multiple of the model timestep")

    if agg:
        if _is_anchored(out_offset):  # e.g. 'ME', 'YE'
            if model_freq_td.total_seconds() > constants.HOURS_PER_DAY * constants.SECONDS_PER_HOUR:
                raise NotImplementedError(
                    "Aggregation of gridded outputs with anchored offsets "
                    "not supported for timesteps > 1d"
                )

            period_end_dates = (
                pd.period_range(
                    start=dates[0],
                    end=dates[-1],
                    freq=out_offset,
                )
                .asfreq(model_freq, how="end")
                .to_timestamp()
            )

            d0 = dates[dates <= period_end_dates[0]][-1]
            write_dates = period_end_dates + (d0 - period_end_dates[0])

            if period_end_dates[0] - write_dates[0] > pd.Timedelta("1d"):
                write_dates = write_dates.delete(0)

            # Keep the last output interval only if it is fully covered (e.g., do not write half
            # months)
            if len(write_dates) > 0 and write_dates[-1] > dates[-1]:
                write_dates = write_dates.delete(-1)
        else:
            write_dates = pd.date_range(
                start=dates[0] + out_freq_td - model_freq_td,
                end=dates[-1],
                freq=out_offset,
            )
    else:
        write_dates = pd.date_range(
            start=dates[0],
            end=dates[-1],
            freq=out_offset,
        )

        if any(
            isinstance(out_offset, o)
            for o in (
                pd.tseries.offsets.YearEnd,
                pd.tseries.offsets.MonthEnd,
            )
        ) and model_freq_td < pd.Timedelta(days=1):
            write_dates += pd.Timedelta(days=1) - model_freq_td

    return write_dates


def _is_anchored(offset: pd.offsets.BaseOffset) -> bool:
    """
    Replicates the is_anchored() method for an offset object, which is deprecated as of pandas
    2.2.0.
    """
    if isinstance(offset, pd.offsets.Tick):
        # https://github.com/pandas-dev/pandas/blob/bdc79c146c2e32f2cab629be240f01658cfb6cc2/pandas/_libs/tslibs/offsets.pyx#L965-L987
        return False
    else:
        # https://github.com/pandas-dev/pandas/blob/bdc79c146c2e32f2cab629be240f01658cfb6cc2/pandas/_libs/tslibs/offsets.pyx#L758-L780
        return offset.n == 1
